##############################################################################
#
# Copyright (C) 2004-2008 Centrify Corporation.
#
# Centrify DirectControl mapper script lock module.
#
# This module create a lock file based on a given file name and lock it.
# If lock can not be created, the constructor will return undef.
#
# On destruction, the lock file will be released and deleted.
#
# There are two ways to create a lock:
#
#   1. specify the file that need to be lock protected. constructor will
#      generate a lock file name based on the given file.
#   2. specify the lock file name directly.
#
# To create a new Lock for file /etc/authorization:
#
#   my $lock = CentrifyDC::GP::Lock->new('/etc/authorization');
#
# To create a new Lock with specified lock file /tmp/authorization.lck:
#
#   my $lock = CentrifyDC::GP::Lock->new('/tmp/authorization.lck', 1);
#
# To get lock file name:
#
#   $lock->lockfile();
#
##############################################################################

use strict;

package CentrifyDC::GP::Lock;
my $VERSION = '1.0';
require 5.000;

use Symbol qw(gensym);
use Fcntl qw(:flock);

sub new($$;$);
sub file($);
sub DESTROY($);



#
# create instance
#
#   $_[0]:  self
#   $_[1]:  file name
#   $_[2]:  file name is lock file name (optional)
#
#   return: self    - successful
#           undef   - failed
#
sub new($$;$)
{
    my ($invocant, $source_file, $is_lockfile) = @_;
    my $class = ref($invocant) || $invocant;

    defined($source_file) or return undef;

    my $file = $source_file;

    if (! $is_lockfile)
    {
        # filename can not contain invalid characters, space, and ~
        $file =~ s%[ ~/\\\[\]:;\|=,\+\*\?<>"]%%mg;
        ($file ne '') or return undef;

        # limit the length of filename
        (length($file) > 200) and $file = substr($file, 0, 200);
        $file = '/tmp/' . $file . '.lock';
    }

    my $fh = gensym;

    open($fh, "> $file") or return undef;
    flock($fh, LOCK_EX|LOCK_NB) or return undef;

    my $self = {
        file    => $file,
        fh      => $fh,
    };

    bless($self, $class);

    return $self;
}

#
# get lock file name
#
#   $_[0]:  self
#
#   return: string  - lock file name
#
sub file($)
{
    return $_[0]->{file};
}

#
# destructor. release lock, close lock file and delete it
#
#   $_[0]:  self
#
sub DESTROY($)
{
    my $self = $_[0];

    flock($self->{fh}, LOCK_UN);
    close($self->{fh});
    unlink($self->{file});
}

1;
